@@ -4,6 +4,8 @@ import numpy as np
 import matplotlib.pyplot as plt
 import matplotlib.gridspec as gridspec
 import time
+import multiprocessing
+import string
 
 class Optimizer:
     """Base class for all optimization methods."""
@@ -13,6 +15,7 @@ class Optimizer:
 
         self.dimensions = None
         self.evaluation_function = None
+        self.number_of_processes = 1
         
         self.objectives = 1
         self.objective_weights = None
@@ -20,37 +23,71 @@ class Optimizer:
         self.constraints = 0
         self.constraint_labels = None
         
+        self.iterations = None
+        self.target_fitness = -np.inf
+        
         self.lb = None
         self.ub = None
         self.verbose = None
         self.report = {}
+
+        self.forward_unique_str = False
+        self._unique_str_list = []
         
-    def init(self):
+    def _init_optimizer(self):
         
-        assert not isinstance(self.objectives, np.unsignedinteger), "optimizer.objectives should be positive integer"
-        assert self.objectives > 0, "An optimizer.objectives should be positive integer"
+        assert not isinstance(self.objectives, np.unsignedinteger), \
+            "optimizer.objectives should be positive integer"
+        assert self.objectives > 0, "optimizer.objectives should be positive integer"
         
-        assert not isinstance(self.constraints, np.unsignedinteger), "optimizer.constraints should be unsigned integer"
+        assert not isinstance(self.constraints, np.unsignedinteger), \
+            "optimizer.constraints should be unsigned integer"
         
         if self.objective_weights is None:
             self.objective_weights = np.ones(self.objectives) / self.objectives   
         else:
-            assert len(self.objective_weights) == self.objectives, "optimizer.objective_weights list should contain number of elements equal to optimizer.objectives"
+            assert len(self.objective_weights) == self.objectives, \
+                "optimizer.objective_weights list should contain number of elements equal to optimizer.objectives"
         
         if self.objective_labels is None:
             self.objective_labels = [f'o_{o}' for o in range(self.objectives)]   
         else:
-            assert len(self.objective_labels) == self.objectives, "optimizer.objective_labels list should contain number of strings equal to optimizer.objectives"
+            assert len(self.objective_labels) == self.objectives, \
+                "optimizer.objective_labels list should contain number of strings equal to optimizer.objectives"
             
         if self.constraint_labels is None:
             self.constraint_labels = [f'c_{c}' for c in range(self.constraints)]   
         else:
-            assert len(self.constraint_labels) == self.constraints, "optimizer.constraint_labels list should contain number of strings equal to optimizer.constraints"
+            assert len(self.constraint_labels) == self.constraints, \
+                "optimizer.constraint_labels list should contain number of strings equal to optimizer.constraints"
             
         self.ET = None
-    
-    def log(self, msg):
-        print('    ' + msg)
+
+        assert isinstance(self.number_of_processes, int) or self.number_of_processes == 'maximum', \
+            "optimizer.number_of_processes should be integer or \'maximum\'"
+        if self.number_of_processes == 'maximum':
+            self.number_of_processes = multiprocessing.cpu_count()
+        if self.number_of_processes > 1:
+            self.log(f'Preparing a multiprocess pool for {self.number_of_processes} processes.')
+            self.pool = multiprocessing.Pool(self.number_of_processes)
+
+
+    def _check_params(self, mandatory_params, optional_params, defined_params):
+        for param in mandatory_params:
+            assert param in defined_params, f'Error: Missing parameter {param}'
+
+        for param in defined_params:
+            if param not in mandatory_params and param not in optional_params:
+                print(f'Warning: Excessive parameter {param}')
+
+    def title(self, msg, line_length=60):
+        t = '-' * line_length
+        t += '\n' + msg.center(line_length) + '\n'
+        t += '-' * line_length
+        print(t)
+
+    def log(self, msg, indent=0):
+        print(' ' * indent + msg)
         
     def tic(self):
         self._time1 = time.time()
@@ -59,17 +96,101 @@ class Optimizer:
 
         s = time.time() - self._time1
         
-        if not silent:        
+        if not silent:
+            if msg == '':
+                msg = 'Elapsed time:'
+
             if s < 1:
-                self.log('{}Elapsed time: {:.3f} miliseconds.'.format('' if msg == '' else msg + '. ', s*1e3))
+                dur = f'{s*1e3:.3f} milliseconds.'
             else:
-                self.log('{}Elapsed time: {:.3f} seconds.'.format('' if msg == '' else msg + '. ', s))
+                dur = f'{s:.3f} seconds.'
+            self.log(f'{msg} {dur}')
             
         return s
 
+
+    def _gen_unique_str(self):
+        while True:
+            _s = ''.join(np.random.choice([c for c in string.ascii_lowercase]) for i in range(16))
+            if _s not in self._unique_str_list:
+                self._unique_str_list.append(_s)
+                return _s
+
+
+    def run(self, seed=None):
+        if None:
+            np.random.seed() # Possibe fix for some seed issues in parallel execution
+        else:
+            np.random.seed(seed)
+
+
+
+    def __getstate__(self):
+        self_dict = self.__dict__.copy()
+        del self_dict['pool']
+        return self_dict
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+
+    def _multiprocess_evaluate(self, candidates):
+
+        if self.forward_unique_str:
+            result = self.pool.starmap(self.evaluation_function, [tuple([c.X, self._gen_unique_str()]) for c in candidates])
+        else:
+            result = self.pool.map(self.evaluation_function, [c.X for c in candidates])
+
+        if self.objectives == 1 and self.constraints == 0:
+            # Fast evaluation
+            for p in range(len(candidates)):
+                candidates[p].f = result[p]
+                candidates[p].O[0] = result[p]
+        else:
+            # Full evaluation
+            for p in range(len(candidates)):
+                for io in range(self.objectives):
+                    candidates[p].O[io] = result[p][io]
+                candidates[p].f = np.dot(candidates[p].O, self.objective_weights)
+                for ic in range(self.constraints):
+                    candidates[p].C[ic] = result[p][self.objectives + ic]
+
+        return candidates
+
+    def collective_evaluation(self, candidates):
+
+        if self.number_of_processes > 1:
+            self._multiprocess_evaluate(candidates)
+        else:
+            if self.forward_unique_str:
+                for p in range(len(candidates)):
+                    candidates[p].evaluate(self._gen_unique_str())
+            else:
+                for p in range(len(candidates)):
+                    candidates[p].evaluate()
+
+    def multiprocess_evaluate_test(self, number_of_candidates, seed=None):
+        # Function for testing purpose only - remove it!
+
+        self._init_optimizer()
+        if None:
+            np.random.seed() # Possibe fix for some seed issues in parallel execution
+        else:
+            np.random.seed(seed)
+
+        candidates = np.array([CandidateState(self) for c in range(number_of_candidates)], dtype=CandidateState)
+
+        for p in range(number_of_candidates):
+            candidates[p].X = np.random.uniform(self.lb, self.ub)
+
+        self.collective_evaluation(candidates)
+
+        return candidates
+
+
+
 class CandidateState:
     """Candidate solution for optimization problem"""
-    
+
     
     def __init__(self, optimizer: Optimizer):
         self._optimizer = optimizer
@@ -80,9 +201,9 @@ class CandidateState:
                     
         # Evaluation
         if self._optimizer.objectives == 1 and self._optimizer.constraints == 0:
-            self.evaluate = self._eval_fast
+            self._evaluate = self._eval_fast
         else:
-            self.evaluate = self._eval_full
+            self._evaluate = self._eval_full
         
         # Comparison operators
         if self._optimizer.objectives == 1 and self._optimizer.constraints == 0:
@@ -152,7 +273,10 @@ class CandidateState:
     # Less-then operator
     def __lt__(self, other):
         return self._lt_fn(self, other)    
-    
+
+    def evaluate(self, s=None):
+        return self._evaluate(s)
+
     @staticmethod
     def _lt_fast(a, b): 
         #print('_lt_fast')
@@ -203,12 +327,14 @@ class CandidateState:
         assert np.size(self.X) == np.size(X), 'Unexpected size of optimization vector X'
         self.X = X
         
-    def _eval_full(self):
+    def _eval_full(self, s=None):
         
         #print('_eval_full')
         #print(self.X.shape)
-        
-        oc = self._optimizer.evaluation_function(self.X)
+        if s is None:
+            oc = self._optimizer.evaluation_function(self.X)
+        else:
+            oc = self._optimizer.evaluation_function(self.X, s)
         #print(oc)
 
         for io in range(self._optimizer.objectives):
@@ -220,14 +346,20 @@ class CandidateState:
             self.C[ic] = oc[self._optimizer.objectives + ic]
 
         #print('_eval_full', self.X[:5], self.f)
+        return self.O, self.C, self.f
         
-    def _eval_fast(self):
-        
-        f = self._optimizer.evaluation_function(self.X)
+    def _eval_fast(self, s=None):
+
+        if s is None:
+            f = self._optimizer.evaluation_function(self.X)
+        else:
+            f = self._optimizer.evaluation_function(self.X, s)
+
         self.f = f
         self.O[0] = f 
         #print('_eval_fast', self.f, type(self.f))
         #input(' >> Press return to continue.')
+        return self.O, self.C, self.f
 
 class OptimizationRecord(CandidateState):
     
